
$main::hilite{case_bin} = 0;
$main::hilite{styl_bin} = 0;
$main::hilite{rexp_bin} = [
	'(\t)',		{ color => 0xffcccc,},
	'(\$#_)',	{ color => 0xff0000,},
	'(#.*$)',	{ color => cl::Gray,},
];

$main::hilite{blok_bin} = [
];
